from copy import deepcopy

from ..node import PipelineNode


class HardFilterNode(PipelineNode):
    def __init__(self, instance_ids: list):
        super().__init__()
        self.instance_ids = instance_ids

    def fit(self):
        return self

    def __get_df_rows(self, dataframe, column, list_to_get):
        return dataframe[dataframe[column].isin(list_to_get)]

    def transform(self):
        self.filtered_table = self.__get_df_rows(
            self.memory.input_table, str(self.memory.id_column), self.instance_ids
        )
        return self

    def pipe_outputs(self):
        self.memory.input_table = self.filtered_table

        def filter_by_id(array_obj):
            return array_obj["id"] in self.instance_ids

        if self.memory.nice_projection is not None:
            self.memory.nice_projection = list(
                filter(filter_by_id, self.memory.nice_projection)
            )
        self.details = {"projection": deepcopy(self.memory.nice_projection)}
        return self.memory

    def export_outputs_js(self):
        return {
            "ids": self.instance_ids,
            "projection": self.memory.nice_projection
            if self.memory.nice_projection is not None
            else [],
        }

    def __repr__(self):
        return "Filter"



from copy import deepcopy
from typing import List

from .filter.hard_id_filter import HardFilterNode
from .lableler.region_based_labler import RegionBasedLabelerNode
from .memory import PipelineMemory
from .node import PipelineNode
from .projection.parametric_umap import ParametricUMAPNode


class Pipeline:
    def __init__(self):
        self.uid = None
        self.global_memory: PipelineMemory = PipelineMemory()  # type: ignore
        self.pipeline: List[PipelineNode] = []
        self.labeler = None

        self.table = None
        self.model = None
        self.id_column = None
        self.name = "default"

    def reset_labeler(self):
        self.labeler = None

    def reset_memory(self):
        if self.global_memory is not None:
            self.global_memory.reset()
        if self.table is not None:
            self.set_table(self.table)
        if self.id_column is not None:
            self.set_id_column(self.id_column)
        if self.model is not None:
            self.set_model(self.model)

    def reset_pipeline(self, up_to_id=""):
        reset_pipeline = []

        up_to_id_exist = len(up_to_id) > 0
        if up_to_id_exist is True:
            for step_node in self.pipeline:
                reset_pipeline.append(step_node)
                if up_to_id == str(step_node.id):
                    break

        self.pipeline = reset_pipeline

    def set_uid(self, uid):
        self.uid = uid

    def set_name(self, name):
        self.name = name
        self.global_memory.name = name

    def set_model(self, model):
        self.model = model
        self.global_memory.model = model

    def set_id_column(self, id_column):
        self.id_column = id_column
        self.global_memory.id_column = id_column

    def set_table(self, table):
        self.table = table
        self.global_memory.global_table = table
        self.global_memory.input_table = table.copy(deep=False)

    def add_filter(self, ids):
        new_node = HardFilterNode(ids)
        new_node.set_memory(self.global_memory)
        new_node.fit()

        self.pipeline.append(new_node)

    def add_umap(self, **kwargs):
        kwargs_umap = {"n_components": 2, "n_epochs": 20, **kwargs}

        new_node = ParametricUMAPNode(**kwargs_umap)
        new_node.set_memory(self.global_memory)
        new_node.fit()

        self.pipeline.append(new_node)

    def add_region_labeler(self, polygon):
        new_node = RegionBasedLabelerNode(polygon)
        new_node.set_memory(self.global_memory)
        new_node.fit()

        self.labeler = new_node

    def run(self, up_to_id=""):
        self.reset_memory()
        js_export = {}
        up_to_id_exist = len(up_to_id) > 0

        curr_node = None
        for step_node in self.pipeline:
            curr_node = step_node.transform()
            self.global_memory = step_node.pipe_outputs()
            if up_to_id_exist is True and up_to_id == str(step_node.id):
                break

        if curr_node is not None:
            js_export = curr_node.json()

        return self.global_memory, js_export

    def run_labeler(self, up_to_id=""):
        js_export = {}
        self.global_memory, js_export = self.run(up_to_id)
        if self.labeler is not None:
            self.global_memory = self.labeler.transform().pipe_outputs()
            js_export = self.labeler.json()

        return self.global_memory, js_export

    def __repr__(self):
        result = f"UID: {self.uid}: "
        for step_node in self.pipeline:
            result += str(step_node) + " -> "
        result += str(self.labeler)
        return result

    def pipeline_repr(self, state=False):
        nodes_repr = []
        final_labeler_json = None
        if state is True:
            self.reset_memory()
            for step_node in self.pipeline:
                step_node.transform()
                self.global_memory = step_node.pipe_outputs()
                js_export = step_node.json()
                nodes_repr.append(deepcopy(js_export))
        else:
            for step_node in self.pipeline:
                nodes_repr.append(step_node.json(state))

        final_labeler_json = None if self.labeler is None else self.labeler.json(state)
        return nodes_repr, final_labeler_json

    def json(self, state=False):
        pipeline, labeler = self.pipeline_repr(state)

        return {
            "name": self.name,
            "uid": self.uid,
            "model": self.model,
            "pipeline": pipeline,
            "labeler": labeler,
        }

    def populated(self):
        return len(self.pipeline) > 0

    def same(self, model: str, uid):
        return model == self.model and uid == self.uid



from .api import (
    distill_function,
    metric_function,
    predict_function,
    transform_function,
    ZenoOptions,
)

__all__ = [
    "predict_function",
    "distill_function",
    "metric_function",
    "transform_function",
    "ZenoOptions",
]



from typing import List

from ..node import PipelineNode


class RegionBasedLabelerNode(PipelineNode):
    def __init__(self, polygon: List[List[float]]):
        super().__init__()
        self.model = polygon

    def point_inside_polygon(self, point: List[float], polygon: List[List[float]]):
        x, y = point
        n = len(polygon)
        inside = False
        p1x, p1y = polygon[0]
        for i in range(n + 1):
            p2x, p2y = polygon[i % n]
            if y > min(p1y, p2y):
                if y <= max(p1y, p2y):
                    if x <= max(p1x, p2x):
                        xinters = float("inf")
                        if p1y != p2y:
                            xinters = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                        if p1x == p2x or x <= xinters:
                            inside = not inside
            p1x, p1y = p2x, p2y
        return inside

    def points_inside_polygon(
        self, points: List[List[float]], polygon: List[List[float]]
    ):
        return [int(self.point_inside_polygon(point, polygon)) for point in points]

    def fit(self):
        return self

    def transform(self):
        assert self.memory.projection is not None, "projection must be present"
        self.labels = self.points_inside_polygon(self.memory.projection, self.model)
        return self

    def pipe_outputs(self):
        self.memory.labels = self.labels
        return self.memory

    def export_outputs_js(self):
        assert self.memory.input_table is not None, "input_table must be present"
        return {
            "labels": self.labels,
            "ids": self.memory.input_table[str(self.memory.id_column)].tolist(),
        }



import numpy as np
from umap.parametric_umap import (  # type: ignore
    load_ParametricUMAP,  # type: ignore
    ParametricUMAP,  # type: ignore
)  # type: ignore

from ...classes import ZenoColumn, ZenoColumnType
from ..node import PipelineNode


class ParametricUMAPNode(PipelineNode):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.model = ParametricUMAP(*args, **kwargs)

    def get_embeddings(self, table, model_name):
        embedding_col = ZenoColumn(
            column_type=ZenoColumnType.EMBEDDING,
            name=model_name,
        )
        embedding_col_name = str(embedding_col)
        embeddings_pd_col = table[embedding_col_name]  # type: ignore
        embeddings = np.stack(embeddings_pd_col.to_numpy())
        return embeddings

    def fit(self):
        embeddings = self.get_embeddings(self.memory.input_table, self.memory.model)
        self.model.fit(embeddings)

        return self

    def transform(self):
        embeddings = self.get_embeddings(self.memory.input_table, self.memory.model)
        self.projections = self.model.transform(embeddings).tolist()

        return self

    def pipe_outputs(self):
        self.memory.projection = [proj for proj in self.projections]
        assert self.memory.input_table is not None, "input_table must be present"
        assert self.memory.id_column is not None, "id_column must be present"
        id_column = self.memory.id_column
        table = self.memory.input_table
        ids = table[str(id_column)].tolist()
        self.memory.nice_projection = self.__package_projection_export(
            self.memory.projection, ids
        )
        return self.memory

    def __package_projection_export(self, projection, instance_ids):
        payload = []
        for proj, instance_id in zip(projection, instance_ids):
            packaged_projection = {"proj": proj, "id": instance_id}
            payload.append(packaged_projection)
        return payload

    def export_outputs_js(self):
        return {"projection": self.memory.nice_projection}

    def save(self, path: str):
        self.model.save(path)

    def load(self, path: str):
        self.model = load_ParametricUMAP(path)

    def __repr__(self):
        return "Projection"



import functools
from dataclasses import dataclass
from typing import Dict


@dataclass
class ZenoOptions:
    """Parameters passed to Zeno test functions.

    Args:
        id_column (str): Column in dataframe with unique identifiers.
        data_column (str): Column in dataframe with either raw data or path to data.
        label_column (str): Column in dataframe with
                            either raw labels or path to labels.
        output_column (str): Column in dataframe with a given
                             model's raw output or path to output
        data_path (str): Path to directory with data files.
        label_path (str): Path to directory with label files.
        output_path (str): Path to directory with a given model's output.
        distill_columns (map[str, str]): Map from distiller name to distill column.
    """

    id_column: str
    data_column: str
    label_column: str
    output_column: str
    distill_columns: Dict[str, str]
    data_path: str
    label_path: str
    output_path: str


def predict_function(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.predict_function = True
    return _wrapper


def distill_function(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.distill_function = True
    return _wrapper


def transform_function(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.transform_function = True
    return _wrapper


def metric_function(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.metric_function = True
    return _wrapper



from enum import IntEnum
from pathlib import Path
from typing import Any, List, Optional, Union

from pydantic import BaseModel


class ZenoColumnType(IntEnum):
    METADATA = 0
    PREDISTILL = 1
    OUTPUT = 2
    EMBEDDING = 3
    POSTDISTILL = 4
    TRANSFORM = 5
    WEAK_LABEL = 6


class ZenoFunctionType(IntEnum):
    PREDICTION = 0
    PREDISTILL = 1
    POSTDISTILL = 2
    TRANSFORM = 3
    METRIC = 4


def to_camel(string):
    components = string.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


class CamelModel(BaseModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True


class ReportPredicate(CamelModel):
    slice_name: str
    metric: str
    transform: str
    operation: str
    value: str
    results: List[int]


class Report(CamelModel):
    name: str
    report_predicates: List[ReportPredicate]


class ZenoFunction(CamelModel):
    name: str
    file_name: Path
    fn_type: ZenoFunctionType


class ZenoColumn(CamelModel):
    column_type: ZenoColumnType
    name: str
    model: Optional[str] = ""
    transform: Optional[str] = ""

    def __str__(self):
        m = ""
        t = ""
        if self.model is not None:
            m = self.model
        if self.transform is not None:
            t = self.transform
        return str(int(self.column_type)) + self.name + m + t


class FilterPredicate(CamelModel):
    column: ZenoColumn
    operation: str
    value: str
    join: str
    group_indicator: Optional[str]


class FilterPredicateGroup(CamelModel):
    predicates: List[Union[FilterPredicate, "FilterPredicateGroup"]]
    join: str


class ZenoSettings(CamelModel):
    view: str
    id_column: ZenoColumn
    label_column: ZenoColumn
    data_column: ZenoColumn
    metadata_columns: List[ZenoColumn]


class ZenoVariables(CamelModel):
    metrics: List[str]
    transforms: List[str]
    models: List[str]
    folders: List[str]


class Slice(CamelModel):
    slice_name: str
    folder: str
    filter_predicates: Optional[FilterPredicateGroup]
    idxs: Optional[List[str]]


class MetricKey(CamelModel):
    sli: Slice
    metric: str
    model: str
    transform: str


class StatusResponse(CamelModel):
    status: str
    done_processing: bool
    complete_columns: List[ZenoColumn]


class PipelineInit(BaseModel):
    model: str
    uid: str = "id:0"


class PipelineReset(BaseModel):
    up_to_id: str = ""


class PipelineIdFilter(BaseModel):
    ids: List[Any]


class PipelineRegionLabeler(BaseModel):
    polygon: List[List[float]]
    name: str = "default"
    up_to_id: str = ""


class PipelineProjection(BaseModel):
    args: dict



import asyncio
import importlib
import json
import os
import sys
from pathlib import Path
from typing import List

import tomli
import uvicorn  # type: ignore
from fastapi import FastAPI, WebSocket
from fastapi.responses import Response
from fastapi.staticfiles import StaticFiles

from .classes import (
    MetricKey,
    PipelineIdFilter,
    PipelineInit,
    PipelineProjection,
    PipelineRegionLabeler,
    PipelineReset,
    Report,
    Slice,
    StatusResponse,
    ZenoColumn,
    ZenoSettings,
    ZenoVariables,
)
from .zeno import Zeno


def main():
    if len(sys.argv) != 2:
        print(
            "ERROR: Zeno take one argument, a configuration TOML file."
            + "{0} arguments were passed.",
            len(sys.argv),
        )
        sys.exit(1)

    args = {}
    try:
        with open(sys.argv[1], "rb") as f:
            args = tomli.load(f)
    except Exception:
        print("ERROR: Failed to read TOML configuration file.")
        sys.exit(1)

    toml_path = os.path.dirname(os.path.abspath(sys.argv[1]))

    if "view" not in args:
        print(
            "ERROR: Must have 'view' entry which is a valid and installed"
            + "Python Zeno view package:"
        )
        sys.exit(1)

    try:
        args["view"] = args["view"].replace("-", "_")
        mod = importlib.import_module(args["view"])
    except ModuleNotFoundError:
        print("ERROR: " + args["view"] + " is not a valid python package")
        sys.exit(1)

    if not mod.path:
        print("ERROR: View does not export a valid path variable.")
        sys.exit(1)

    args["view"] = mod.path

    if "metadata" not in args:
        print("ERROR: Must have 'metadata' entry which must be a CSV or Parquet file.")
        sys.exit(1)
    else:
        args["metadata"] = Path(
            os.path.realpath(os.path.join(toml_path, args["metadata"]))
        )

    if "tests" not in args or not os.path.exists(args["tests"]):
        print("WARNING: No 'tests' directory found.")
        args["tests"] = None
    else:
        args["tests"] = Path(os.path.realpath(os.path.join(toml_path, args["tests"])))

    if "models" not in args or len(args["models"]) < 1:
        print("WARNING: No 'models' found.")
        args["models"] = []
    else:
        if Path(os.path.realpath(os.path.join(toml_path, args["models"][0]))).exists():
            args["models"] = [
                Path(os.path.realpath(os.path.join(toml_path, m)))
                for m in args["models"]
            ]

    if "data_path" not in args:
        args["data_path"] = ""
    else:
        args["data_path"] = Path(
            os.path.realpath(os.path.join(toml_path, args["data_path"]))
        )

    if "label_path" not in args:
        args["label_path"] = ""
    else:
        args["label_path"] = Path(
            os.path.realpath(os.path.join(toml_path, args["label_path"]))
        )

    if "id_column" not in args:
        if "data_column" in args:
            args["id_column"] = args["data_column"]
        else:
            print(
                "ERROR: Must have 'id_column' referencing a column with unique IDs",
                "if no data_column is specified.",
            )

    if "data_column" not in args:
        args["data_column"] = ""

    if "label_column" not in args:
        args["label_column"] = "label"

    if "port" not in args:
        args["port"] = 8000

    if "batch_size" not in args:
        args["batch_size"] = 1

    if "cache_path" not in args:
        args["cache_path"] = Path(
            os.path.realpath(os.path.join(toml_path, "./.zeno_cache/"))
        )
    else:
        args["cache_path"] = Path(
            os.path.realpath(os.path.join(toml_path, args["cache_path"]))
        )

    run_zeno(args)


def run_zeno(args):
    zeno = Zeno(
        metadata_path=args["metadata"],
        tests=args["tests"],
        models=args["models"],
        batch_size=args["batch_size"],
        id_column=args["id_column"],
        data_column=args["data_column"],
        label_column=args["label_column"],
        data_path=args["data_path"],
        label_path=args["label_path"],
        cache_path=args["cache_path"],
    )

    zeno.start_processing()

    app = FastAPI(title="Frontend API")
    api_app = FastAPI(title="Backend API")

    if args["data_path"] != "":
        app.mount("/data", StaticFiles(directory=args["data_path"]), name="static")
    if args["label_path"] != "":
        app.mount("/labels", StaticFiles(directory=args["label_path"]), name="static")

    app.mount(
        "/cache",
        StaticFiles(directory=args["cache_path"]),
        name="cache",
    )
    app.mount(
        "/view",
        StaticFiles(directory=args["view"]),
        name="view",
    )
    app.mount("/api", api_app)
    app.mount(
        "/",
        StaticFiles(
            directory=os.path.dirname(os.path.realpath(__file__)) + "/frontend",
            html=True,
        ),
        name="base",
    )

    @api_app.get("/settings", response_model=ZenoSettings)
    def get_settings():
        return ZenoSettings(
            view=args["view"],
            id_column=zeno.id_column,
            label_column=zeno.label_column,
            data_column=zeno.data_column,
            metadata_columns=zeno.columns,
        )

    @api_app.get("/initialize", response_model=ZenoVariables)
    def get_initial_info():
        return ZenoVariables(
            metrics=list(zeno.metric_functions.keys()),
            transforms=list(zeno.transform_functions.keys()),
            models=[str(n) for n in zeno.model_names],
            folders=zeno.folders,
        )

    @api_app.get("/slices")
    def get_slices():
        return json.dumps(zeno.get_slices())

    @api_app.get("/reports")
    def get_reports():
        return json.dumps(zeno.get_reports())

    @api_app.post("/results")
    def get_results(reqs: List[MetricKey]):
        return json.dumps(zeno.get_results(reqs))

    @api_app.post("/table")
    def get_table(columns: List[ZenoColumn]):
        return Response(zeno.get_table(columns))

    @api_app.post("/set-folders")
    def set_folders(folders: List[str]):
        zeno.set_folders(folders)

    @api_app.post("/set-slices")
    def set_slices(slices: List[Slice]):
        zeno.set_slices(slices)

    @api_app.post("/set-reports")
    def update_reports(reqs: List[Report]):
        zeno.set_reports(reqs)

    @api_app.post("/pipe/reset")
    def reset_pipeline(req: PipelineReset):
        zeno.reset_pipeline(req.up_to_id)
        return json.dumps({"status": True})

    @api_app.post("/pipe/load")
    def load_pipeline(req: PipelineInit):
        js_export = zeno.load_pipeline(req.model, req.uid)
        return json.dumps({"status": True, "data": js_export})

    @api_app.post("/pipe/init")
    def init_pipeline(req: PipelineInit):
        zeno.init_pipeline(req.model, req.uid)
        return json.dumps({"status": True})

    @api_app.post("/pipe/id-filter")
    def filter_pipeline(req: PipelineIdFilter):
        js_export = zeno.add_id_filter_pipeline(req.ids)
        return json.dumps({"status": True, "data": js_export})

    @api_app.post("/pipe/umap")
    def umap_pipeline(req: PipelineProjection):
        js_export = zeno.add_umap_pipeline(req.args)
        return json.dumps({"status": True, "data": js_export})

    @api_app.post("/pipe/region-labeler")
    def region_labeler_pipeline(req: PipelineRegionLabeler):
        js_export = zeno.add_region_labeler_pipeline(
            req.polygon, req.name, req.up_to_id
        )
        return json.dumps({"status": True, "data": js_export})

    @api_app.get("/pipe/graph")
    def get_pipeline_json():
        js_export = zeno.get_json_pipeline()
        return json.dumps({"status": True, "data": js_export})

    @api_app.websocket("/status")
    async def results_websocket(websocket: WebSocket):
        await websocket.accept()
        previous_status = ""
        while True:
            await asyncio.sleep(1)
            if zeno.status != previous_status:
                previous_status = zeno.status
                await websocket.send_json(
                    StatusResponse(
                        status=zeno.status,
                        done_processing=zeno.done_processing,
                        complete_columns=zeno.complete_columns,
                    ).json(by_alias=True)
                )

    uvicorn.run(app, host="localhost", port=args["port"])  # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeno',
 'zeno.pipeline',
 'zeno.pipeline.filter',
 'zeno.pipeline.lableler',
 'zeno.pipeline.projection']

package_data = \
{'': ['*'],
 'zeno': ['frontend/*', 'frontend/build/*', 'frontend/build/assets/*']}

install_requires = \
['fastapi>=0.75,<0.80',
 'pandas>=1.4.0,<2.0.0',
 'pyarrow>=7.0.0,<8.0.0',
 'tensorflow>=2.9.0,<3.0.0',
 'tomli>=2.0.1,<3.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'umap-learn>=0.5.3,<0.6.0',
 'uvicorn>=0.17.5,<0.19.0',
 'websockets>=10.2,<11.0']

entry_points = \
{'console_scripts': ['zeno = zeno.runner:main']}

setup_kwargs = {
    'name': 'zenoml',
    'version': '0.0.6',
    'description': 'Behavioral Evaluation Framework for Machine Learning',
    'long_description': '<img src="./frontend/public/zeno.png" width="400px"/>\n\n[![PyPI version](https://badge.fury.io/py/zenoml.svg)](https://badge.fury.io/py/zenoml)\n![Github Actions CI tests](https://github.com/cabreraalex/zeno/actions/workflows/test.yml/badge.svg)\n![Github Actions Docs build](https://github.com/cabreraalex/zeno/actions/workflows/book.yml/badge.svg)\n[![MIT license](https://img.shields.io/badge/License-MIT-blue.svg)](https://lbesson.mit-license.org/)\n\nZeno is a general-purpose framework for evaluating machine learning models.\nIt combines a **Python API** with an **interactive UI** to allow users to discover, explore, and analyze the performance of their models across diverse use cases.\nZeno can be used for any data type or task with _modular views_ for everything from object detection to audio transcription.\n\n## Getting Started\n\nCheck out the quickstart tutorial and API reference to get started with Zeno:\n\n- [Introduction](http://zenoml.com/docs/intro/) - Learn more about Zeno.\n- [Getting Started](http://zenoml.com/docs/intro/get_started) - Setup Zeno with your own data and models.\n- [CIFAR-10 Example](http://zenoml.com/docs/intro/cifar) - Learn how to use Zeno with a CIFAR-10 classification example.\n- [Documentation & API](http://zenoml.com/docs/intro/) - Full documentation and API reference.\n',
    'author': 'Ángel Alexander Cabrera',
    'author_email': 'alex.cabrera@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://dig.cmu.edu/zeno/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from contextlib import contextmanager
import dataclasses
import os
import pickle
from importlib import util
from inspect import getsource
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
import pyarrow as pa  # type: ignore
from tqdm import trange  # type: ignore

from .api import ZenoOptions
from .classes import ZenoColumn, ZenoColumnType, ZenoFunction


@contextmanager
def add_to_path(p):
    import sys

    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def get_arrow_bytes(df, id_col):
    df[id_col] = df.index
    df = df.infer_objects()
    d = dict.fromkeys(df.select_dtypes(np.int64).columns, np.int32)
    df = df.astype(d)
    df_arrow = pa.Table.from_pandas(df)
    buf = pa.BufferOutputStream()
    with pa.ipc.new_file(buf, df_arrow.schema) as writer:
        writer.write_table(df_arrow)
    bs = bytes(buf.getvalue())
    return bs


def read_pickle(file_name: str, cache_path: Path, default):
    try:
        with open(os.path.join(cache_path, file_name), "rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        return default


def load_series(df, col_name, save_path):
    try:
        df.loc[:, col_name] = pd.read_pickle(save_path)
    except FileNotFoundError:
        df.loc[:, col_name] = pd.Series([pd.NA] * df.shape[0], index=df.index)


def get_function(fn: ZenoFunction):
    # To allow relative imports in test files,
    # add their directory to path temporarily.
    with add_to_path(os.path.dirname(os.path.abspath(fn.file_name))):
        spec = util.spec_from_file_location("module.name", fn.file_name)
        test_module = util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(test_module)  # type: ignore
        return getattr(test_module, fn.name)  # type: ignore


def predistill_data(
    preprocessor: ZenoFunction,
    column: ZenoColumn,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> Tuple[ZenoColumn, pd.Series]:
    fn = get_function(preprocessor)
    col_hash = str(column)
    col = df[col_hash]

    save_path = Path(cache_path, col_hash + ".pickle")
    to_predict_indices = col.loc[pd.isna(col)].index

    if len(to_predict_indices) > 0:
        if len(to_predict_indices) < batch_size:
            out = fn(df.loc[to_predict_indices], options)
            col.loc[to_predict_indices] = out
            col.to_pickle(str(save_path))
        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="preprocessing " + preprocessor.name,
                position=pos,
            ):
                out = fn(df.loc[to_predict_indices[i : i + batch_size]], options)
                col.loc[to_predict_indices[i : i + batch_size]] = out
                col.to_pickle(str(save_path))
    return (column, col)


def transform_data(
    transform: ZenoFunction,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> Tuple[ZenoColumn, pd.Series]:
    transform_fn = get_function(transform)
    transform_col = ZenoColumn(
        column_type=ZenoColumnType.TRANSFORM,
        name=transform.name,
    )
    col_hash = str(transform_col)
    col = df[col_hash]

    save_path = Path(cache_path, col_hash + ".pickle")
    src = getsource(transform_fn)
    transform_cache_path = ""
    if "output_path" in src:
        transform_cache_path = os.path.join(cache_path, col_hash)
        os.makedirs(transform_cache_path, exist_ok=True)

    to_transform_indices = col.loc[pd.isna(col)].index

    if len(to_transform_indices) > 0:
        if len(to_transform_indices) < batch_size:
            out = transform_fn(
                df.loc[to_transform_indices],
                dataclasses.replace(options, output_path=transform_cache_path),
            )
            col.loc[to_transform_indices] = out
            col.to_pickle(str(save_path))
        else:
            for i in trange(
                0,
                len(to_transform_indices),
                batch_size,
                desc="transforming " + transform.name,
                position=pos,
            ):
                out = transform_fn(
                    df.loc[to_transform_indices[i : i + batch_size]],
                    dataclasses.replace(options, output_path=transform_cache_path),
                )
                col.loc[to_transform_indices[i : i + batch_size]] = out
                col.to_pickle(str(save_path))
    return (transform_col, col)


def run_inference(
    model_loader: ZenoFunction,
    options: ZenoOptions,
    transform: str,
    model_path: str,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> Tuple[ZenoColumn, ZenoColumn, pd.Series, pd.Series]:
    model_loader_fn = get_function(model_loader)
    model_name = os.path.basename(model_path).split(".")[0]

    model_col_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT,
        name=model_name,
        transform=transform,
    )
    embedding_col_obj = ZenoColumn(
        column_type=ZenoColumnType.EMBEDDING,
        name=model_name,
        transform=transform,
    )
    model_hash = str(model_col_obj)
    embedding_hash = str(embedding_col_obj)
    model_col = df[model_hash]
    embedding_col = df[embedding_hash]

    model_save_path = Path(cache_path, model_hash + ".pickle")
    embedding_save_path = Path(cache_path, embedding_hash + ".pickle")

    to_predict_indices = model_col.loc[pd.isna(model_col)].index

    # If transform, pass transform data
    if len(transform) != 0:
        transform_data_col = ZenoColumn(
            column_type=ZenoColumnType.TRANSFORM,
            name=transform,
        )
        transform_col_name = str(transform_data_col)
        transform_path = os.path.join(cache_path, transform_col_name)
        options = dataclasses.replace(
            options, data_path=transform_path, data_column=transform_col_name
        )

    if len(to_predict_indices) > 0:
        fn = model_loader_fn(model_path)
        if len(to_predict_indices) < batch_size:
            # Make output folder if function uses output_path
            src = getsource(fn)
            if "output_path" in src:
                file_cache_path = os.path.join(cache_path, model_hash)
                os.makedirs(file_cache_path, exist_ok=True)
                options = dataclasses.replace(options, output_path=file_cache_path)

            out = fn(df.loc[to_predict_indices], options)

            # Check if we also get embedding
            if type(out) == tuple and len(out) == 2:
                for i, idx in enumerate(to_predict_indices):
                    model_col.at[idx] = out[0][i]
                    embedding_col.at[idx] = out[1][i]
                embedding_col.to_pickle(str(embedding_save_path))
                out = out[0]
            else:
                model_col[to_predict_indices] = out

            model_col.to_pickle(str(model_save_path))

        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="Inference on " + model_name + " with " + transform,
                position=pos,
            ):

                # Make output folder if function uses output_path
                src = getsource(fn)
                if "output_path" in src:
                    file_cache_path = os.path.join(cache_path, model_hash)
                    os.makedirs(file_cache_path, exist_ok=True)
                    options = dataclasses.replace(options, output_path=file_cache_path)

                out = fn(df.loc[to_predict_indices[i : i + batch_size]], options)

                # Check if we also get embedding
                if type(out) == tuple and len(out) == 2:
                    for i, idx in enumerate(to_predict_indices[i : i + batch_size]):
                        model_col.at[idx] = out[0][i]
                        embedding_col.at[idx] = out[1][i]
                    embedding_col.to_pickle(str(embedding_save_path))
                    out = out[0]
                else:
                    model_col[to_predict_indices[i : i + batch_size]] = out

                model_col.to_pickle(str(model_save_path))

    return (model_col_obj, embedding_col_obj, model_col, embedding_col)


def postdistill_data(
    postprocessor: ZenoFunction,
    model: str,
    transform: str,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
):
    postprocessor_fn = get_function(postprocessor)
    col_obj = ZenoColumn(
        column_type=ZenoColumnType.POSTDISTILL,
        name=postprocessor.name,
        model=model,
        transform=transform,
    )
    col_hash = str(col_obj)
    col = df[col_hash]

    output_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT,
        name=model,
        transform=transform,
    )
    output_hash = str(output_obj)

    save_path = Path(cache_path, col_hash + ".pickle")

    to_predict_indices = col.loc[pd.isna(col)].index

    local_options = dataclasses.replace(
        options,
        output_column=output_hash,
        output_path=os.path.join(cache_path, output_hash),
    )
    if len(to_predict_indices) > 0:
        if len(to_predict_indices) < batch_size:
            out = postprocessor_fn(df.loc[to_predict_indices], local_options)
            col.loc[to_predict_indices] = out
            col.to_pickle(str(save_path))
        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="postprocessing "
                + postprocessor.name
                + " on "
                + model
                + " with "
                + transform,
                position=pos,
            ):
                out = postprocessor_fn(
                    df.loc[to_predict_indices[i : i + batch_size]], local_options
                )
                col.loc[to_predict_indices[i : i + batch_size]] = out
                col.to_pickle(str(save_path))
    return (col_obj, col)



from typing import List, Union

from pandas import DataFrame

from ..classes import ZenoColumn


class PipelineMemory:
    def __init__(
        self,
        projection: Union[List[List[float]], None] = None,
        nice_projection: Union[List[dict], None] = None,
        input_table: Union[DataFrame, None] = None,
        global_table: Union[DataFrame, None] = None,
        name: Union[str, None] = None,
        id_column: Union[ZenoColumn, None] = None,
        model: Union[str, None] = None,
        labels: Union[List[Union[int, bool]], None] = None,
    ):
        self.projection = projection
        self.nice_projection = nice_projection
        self.input_table = input_table
        self.global_table = global_table
        self.name = name
        self.id_column = id_column
        self.model = model
        self.labels = labels

    def reset(self):
        self.projection = None
        self.nice_projection = None
        self.input_table = None
        self.global_table = None
        self.name = None
        self.id_column = None
        self.model = None
        self.labels = None



from .memory import PipelineMemory


class PipelineNode:
    def __init__(self):
        self.id = NodeId()
        self.state = {}
        self.memory: PipelineMemory

    def fit(self) -> "PipelineNode":
        return self

    def transform(self) -> "PipelineNode":
        return self

    def pipe_outputs(self) -> PipelineMemory:
        return self.memory

    def export_outputs_js(self) -> dict:
        return {}

    def save(self, path: str):
        pass

    def load(self, path: str):
        pass

    def get_memory(self):
        return self.memory

    def set_memory(self, memory: PipelineMemory) -> "PipelineNode":
        self.memory = memory
        return self

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def json(self, state=True):
        return {
            "type": self.__repr__(),
            "id": str(self.id),
            "state": self.export_outputs_js() if state is True else None,
        }


class NodeId:
    def __init__(self):
        self.id = None
        self.generate_random_id()

    def generate_random_id(self):
        from uuid import uuid4

        self.id = uuid4()

    def __repr__(self) -> str:
        return str(self.id)

    def __str__(self) -> str:
        return self.__repr__()



import asyncio
import dataclasses
import logging
import os
import pickle
import sys
import threading
from importlib import util
from inspect import getmembers, getsource, isfunction
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd

from .api import ZenoOptions
from .classes import (
    MetricKey,
    Report,
    Slice,
    ZenoColumn,
    ZenoColumnType,
    ZenoFunction,
    ZenoFunctionType,
)
from .pipeline.pipeline import Pipeline
from .util import (
    add_to_path,
    get_arrow_bytes,
    get_function,
    load_series,
    postdistill_data,
    predistill_data,
    read_pickle,
    run_inference,
    transform_data,
)


class Zeno(object):
    def __init__(
        self,
        metadata_path: Path,
        tests: Path,
        models: List[Path],
        batch_size,
        id_column,
        data_column,
        label_column,
        data_path,
        label_path,
        cache_path: Path,
    ):
        logging.basicConfig(level=logging.INFO)

        self.pipeline = Pipeline()

        self.tests = tests
        self.batch_size = batch_size
        self.id_column = ZenoColumn(column_type=ZenoColumnType.METADATA, name=id_column)
        self.data_column = ZenoColumn(
            column_type=ZenoColumnType.METADATA, name=data_column
        )
        self.label_column = ZenoColumn(
            column_type=ZenoColumnType.METADATA, name=label_column
        )
        self.data_path = data_path
        self.label_path = label_path

        # Options passed to Zeno functions.
        self.zeno_options = ZenoOptions(
            id_column=str(self.id_column),
            data_column=str(self.data_column),
            label_column=str(self.label_column),
            distill_columns=dict(),
            data_path=self.data_path,
            label_path=self.label_path,
            output_column="",
            output_path="",
        )

        if models and os.path.isdir(models[0]):
            self.model_paths = [
                os.path.join(models[0], m) for m in os.listdir(models[0])
            ]
        else:
            self.model_paths = models  # type: ignore
        self.model_names = [os.path.basename(p).split(".")[0] for p in self.model_paths]

        self.status: str = "Initializing"

        self.predict_function: ZenoFunction = ZenoFunction(
            name="", file_name=Path(""), fn_type=ZenoFunctionType.PREDICTION
        )
        self.distill_functions: Dict[str, ZenoFunction] = {}
        self.metric_functions: Dict[str, ZenoFunction] = {}
        self.transform_functions: Dict[str, ZenoFunction] = {}

        # Read metadata as Pandas for slicing
        if metadata_path.suffix == ".csv":
            self.df = pd.read_csv(metadata_path)
        elif metadata_path.suffix == ".parquet":
            self.df = pd.read_parquet(metadata_path)
        else:
            logging.error(
                "Extension of " + metadata_path.suffix + " not one of .csv or .parquet"
            )
            sys.exit(1)

        # TODO: figure out if this breaks for big integers. Need to do this for
        # frontend bigint issues.
        d = dict.fromkeys(
            self.df.select_dtypes(np.int64).columns, np.int32  # type: ignore
        )
        self.df = self.df.astype(d)  # type: ignore

        self.metadata_name = os.path.basename(metadata_path).split(".")[0]
        self.cache_path = cache_path
        os.makedirs(self.cache_path, exist_ok=True)

        self.folders: List[str] = read_pickle("folders.pickle", self.cache_path, [])
        self.reports: List[Report] = read_pickle("reports.pickle", self.cache_path, [])
        self.slices: Dict[str, Slice] = read_pickle(
            "slices.pickle", self.cache_path, {}
        )

        self.df = self.df.rename(columns=lambda x: "0" + str(x))
        self.df[str(self.id_column)].astype(str)
        self.df.set_index(str(self.id_column), inplace=True)
        self.df[str(self.id_column)] = self.df.index

        self.columns: List[ZenoColumn] = []
        self.complete_columns: List[ZenoColumn] = []

        self.done_processing = False

    def start_processing(self):
        """Parse testing files, distill, and run inference."""

        # Get all Zeno core columns before processing.
        zeno_cols: List[ZenoColumn] = []
        for metadata_col in self.df.columns:
            col = ZenoColumn(column_type=ZenoColumnType.METADATA, name=metadata_col[1:])
            zeno_cols.append(col)
            self.complete_columns.append(col)

        # Get Zeno test functions.
        if not self.tests:
            self.columns = zeno_cols
            self.done_processing = True
            self.status = "Done processing"
            return

        # Add directory with tests to path for relative imports.
        for f in list(self.tests.rglob("*.py")):
            self.__parse_testing_file(f)

        for fn in self.distill_functions.values():
            if fn.fn_type == ZenoFunctionType.PREDISTILL:
                zeno_cols.append(
                    ZenoColumn(column_type=ZenoColumnType.PREDISTILL, name=fn.name)
                )
            else:
                for m in self.model_names:
                    zeno_cols.append(
                        ZenoColumn(
                            column_type=ZenoColumnType.POSTDISTILL,
                            name=fn.name,
                            model=m,
                        )
                    )

        self.columns = zeno_cols

        self.__thread = threading.Thread(target=asyncio.run, args=(self.__process(),))
        self.__thread.start()

    def __parse_testing_file(self, test_file: Path):
        # To allow relative imports in test files,
        # add their directory to path temporarily.
        with add_to_path(os.path.dirname(os.path.abspath(test_file))):
            spec = util.spec_from_file_location(str(test_file), test_file)
            test_module = util.module_from_spec(spec)  # type: ignore
            spec.loader.exec_module(test_module)  # type: ignore

        for func_name, func in getmembers(test_module):
            if isfunction(func):
                if hasattr(func, "predict_function"):
                    if self.predict_function.name == "":
                        self.predict_function = ZenoFunction(
                            name=func_name,
                            file_name=test_file,
                            fn_type=ZenoFunctionType.PREDICTION,
                        )
                    else:
                        logging.error(
                            "Multiple prediction_functions found, can only have one"
                        )
                        sys.exit(1)
                if hasattr(func, "distill_function"):
                    src = getsource(func)
                    self.distill_functions[func_name] = ZenoFunction(
                        name=func_name,
                        file_name=test_file,
                        fn_type=(
                            ZenoFunctionType.POSTDISTILL
                            if "output_column" in src
                            else ZenoFunctionType.PREDISTILL
                        ),
                    )
                if hasattr(func, "transform_function"):
                    self.transform_functions[func_name] = ZenoFunction(
                        name=func_name,
                        file_name=test_file,
                        fn_type=ZenoFunctionType.TRANSFORM,
                    )
                if hasattr(func, "metric_function"):
                    self.metric_functions[func_name] = ZenoFunction(
                        name=func_name,
                        file_name=test_file,
                        fn_type=ZenoFunctionType.METRIC,
                    )

    async def __process(self):
        self.status = "Running distill functions"
        self.__predistill()
        self.status = "Running transform functions"
        self.__transform()

        for transform in [
            *self.transform_functions.values(),
            ZenoFunction(
                name="", file_name=Path(""), fn_type=ZenoFunctionType.TRANSFORM
            ),
        ]:
            self.__inference_and_postdistill(transform)

        self.done_processing = True

    def __transform(self):
        """Run transform functions."""
        transform_to_run: List[ZenoFunction] = []
        for transform_function in self.transform_functions.values():
            transform_column = ZenoColumn(
                column_type=ZenoColumnType.TRANSFORM, name=transform_function.name
            )
            save_path = Path(self.cache_path, str(transform_column) + ".pickle")

            load_series(self.df, str(transform_column), save_path)

            if self.df[str(transform_column)].isnull().any():
                transform_to_run.append(transform_function)
            else:
                self.complete_columns.append(transform_column)

        if len(transform_to_run) > 0:
            with Pool() as pool:
                transform_outputs = pool.starmap(
                    transform_data,
                    [
                        [
                            transform_fn,
                            self.zeno_options,
                            self.cache_path,
                            self.df,
                            self.batch_size,
                            i,
                        ]
                        for i, transform_fn in enumerate(transform_to_run)
                    ],
                )
                for out in transform_outputs:
                    self.df.loc[:, str(out[0])] = out[1]
                    self.complete_columns.append(out[0])

    def __predistill(self):
        """Run distilling functions not dependent on model outputs."""
        # Check if we need to preprocess since Pool is expensive
        predistill_to_run: List[ZenoColumn] = []
        for predistill_column in [
            c for c in self.columns if c.column_type == ZenoColumnType.PREDISTILL
        ]:
            save_path = Path(self.cache_path, str(predistill_column) + ".pickle")

            load_series(self.df, str(predistill_column), save_path)

            if self.df[str(predistill_column)].isnull().any():
                predistill_to_run.append(predistill_column)
            else:
                self.complete_columns.append(predistill_column)

        if len(predistill_to_run) > 0:
            with Pool() as pool:
                predistill_outputs = pool.starmap(
                    predistill_data,
                    [
                        [
                            self.distill_functions[col.name],
                            col,
                            self.zeno_options,
                            self.cache_path,
                            self.df,
                            self.batch_size,
                            i,
                        ]
                        for i, col in enumerate(predistill_to_run)
                    ],
                )
                for out in predistill_outputs:
                    self.df.loc[:, str(out[0])] = out[1]
                    self.complete_columns.append(out[0])

    def __inference_and_postdistill(self, transform: ZenoFunction):
        # Check if we need to run inference since Pool is expensive
        models_to_run = []
        for model_path in self.model_paths:
            model_name = os.path.basename(model_path).split(".")[0]
            model_column = ZenoColumn(
                column_type=ZenoColumnType.OUTPUT,
                name=model_name,
                transform=transform.name,
            )
            embedding_column = ZenoColumn(
                column_type=ZenoColumnType.EMBEDDING,
                name=model_name,
                transform=transform.name,
            )
            model_hash = str(model_column)
            embedding_hash = str(embedding_column)

            model_save_path = Path(self.cache_path, model_hash + ".pickle")
            embedding_save_path = Path(self.cache_path, embedding_hash + ".pickle")

            load_series(self.df, model_hash, model_save_path)
            load_series(self.df, embedding_hash, embedding_save_path)

            if (
                self.df[model_hash].isnull().any()
                or self.df[embedding_hash].isnull().any()
            ):
                models_to_run.append(model_path)
            else:
                self.complete_columns.append(model_column)

        self.status = "Running inference"
        if len(models_to_run) > 0:
            with Pool() as pool:
                inference_outputs = pool.starmap(
                    run_inference,
                    [
                        [
                            self.predict_function,
                            self.zeno_options,
                            transform.name,
                            m,
                            self.cache_path,
                            self.df,
                            self.batch_size,
                            i,
                        ]
                        for i, m in enumerate(models_to_run)
                    ],
                )
                for out in inference_outputs:
                    self.df.loc[:, str(out[0])] = out[2]
                    self.df.loc[:, str(out[1])] = out[3]
                    self.complete_columns.append(out[0])

        self.status = "Done running inference"

        # Check if we need to run postprocessing since Pool is expensive
        postdistill_to_run: List[ZenoColumn] = []
        for postdistill_column in [
            c for c in self.columns if c.column_type == ZenoColumnType.POSTDISTILL
        ]:
            col_name = postdistill_column.copy(
                update={"model": postdistill_column.model, "transform": transform.name}
            )
            col_hash = str(col_name)
            save_path = Path(self.cache_path, col_hash + ".pickle")

            load_series(self.df, col_hash, save_path)

            if self.df[col_hash].isnull().any():
                postdistill_to_run.append(col_name)
            else:
                self.complete_columns.append(col_name)

        self.status = "Running postprocessing"
        if len(postdistill_to_run) > 0:
            with Pool() as pool:
                post_outputs = pool.starmap(
                    postdistill_data,
                    [
                        [
                            self.distill_functions[e.name],
                            e.model,
                            transform.name,
                            self.zeno_options,
                            self.cache_path,
                            self.df,
                            self.batch_size,
                            i,
                        ]
                        for i, e in enumerate(postdistill_to_run)
                    ],
                )
                for out in post_outputs:
                    self.df.loc[:, str(out[0])] = out[1]  # type: ignore
                    self.complete_columns.append(out[0])

        self.status = "Done running postprocessing"

    def get_results(self, requests: List[MetricKey]):
        """Calculate result for each requested combination."""

        return_metrics = []
        for metric_key in requests:
            return_metrics.append(self.calculate_metrics(metric_key))

        return return_metrics

    def calculate_metrics(self, metric_key: MetricKey):
        if not self.done_processing:
            return

        sli = metric_key.sli
        output_col = ZenoColumn(
            column_type=ZenoColumnType.OUTPUT,
            name=metric_key.model,
            transform=metric_key.transform,
        )
        output_hash = str(output_col)

        distill_fns = [
            c
            for c in self.columns
            if (
                c.column_type == ZenoColumnType.PREDISTILL
                or c.column_type == ZenoColumnType.POSTDISTILL
            )
            and c.transform == metric_key.transform
            and c.model == metric_key.model
        ]

        local_ops = dataclasses.replace(
            self.zeno_options,
            output_column=output_hash,
            output_path=os.path.join(self.cache_path, output_hash),
            distill_columns=dict(
                zip([c.name for c in distill_fns], [str(c) for c in distill_fns])
            ),
        )
        if sli.idxs and len(sli.idxs) > 0:
            metric_func = get_function(self.metric_functions[metric_key.metric])
            result_metric = metric_func(self.df.loc[sli.idxs], local_ops)
        else:
            result_metric = None

        return result_metric

    def reset_pipeline(self, up_to_id=""):
        self.pipeline.reset_memory()
        self.pipeline.reset_pipeline(up_to_id)
        self.pipeline.reset_labeler()

    def init_pipeline(self, model: str, uid: str = "0"):
        self.pipeline.set_uid(uid)
        self.pipeline.set_model(model)
        self.pipeline.set_id_column(self.id_column)
        self.pipeline.set_table(self.df)

    def add_id_filter_pipeline(self, ids: List[str]):
        self.pipeline.add_filter(ids)
        output, js_export = self.pipeline.run()
        return js_export

    def add_umap_pipeline(self, user_specified_args: dict):
        self.pipeline.add_umap(**user_specified_args)
        output, js_export = self.pipeline.run()
        return js_export

    def get_json_pipeline(self):
        return self.pipeline.json()

    def load_pipeline(self, model: str, uid: str = "0"):
        if self.pipeline.populated() and self.pipeline.same(model, uid):
            return self.pipeline.json(state=True)
        else:
            # within the init I can do cache stuff
            self.init_pipeline(model, uid)
            self.reset_pipeline()
            return None

    def add_region_labeler_pipeline(self, polygon, name, up_to_id=""):
        self.pipeline.set_name(name)
        self.pipeline.add_region_labeler(polygon)
        output, js_export = self.pipeline.run_labeler(up_to_id)
        col_name = self.table_weak_labels(js_export["state"])
        return {"export": js_export["state"], "col_name": col_name}

    def table_weak_labels(self, js_export):
        col_kwargs = {
            "column_type": ZenoColumnType.WEAK_LABEL,
            "name": self.pipeline.name,
            "model": self.pipeline.model,
            "transform": "",
        }
        new_column_labels = ZenoColumn(
            **col_kwargs,
        )
        new_column_labels_hash = str(new_column_labels)

        ids = js_export["ids"]
        labels = js_export["labels"]
        assert len(ids) == len(labels), "Yo your're algo is wrong buddy"
        self.df.loc[:, new_column_labels_hash] = 0
        self.df.loc[ids, new_column_labels_hash] = labels

        self.complete_columns.append(new_column_labels)
        self.columns.append(new_column_labels)
        self.status = f"Done running weak labeler {new_column_labels_hash}"

        return col_kwargs

    def get_table(self, columns):
        """Get the metadata DataFrame for a given slice.

        Returns:
            bytes: Arrow-encoded table of slice metadata
        """
        return get_arrow_bytes(self.df[[str(c) for c in columns]], str(self.id_column))

    def get_slices(self):
        return [s.dict(by_alias=True) for s in self.slices.values()]

    def get_reports(self):
        return [r.dict(by_alias=True) for r in self.reports]

    def set_folders(self, folders):
        self.folders = folders
        with open(os.path.join(self.cache_path, "folders.pickle"), "wb") as f:
            pickle.dump(self.folders, f)

    def set_reports(self, reports: List[Report]):
        self.reports = reports
        with open(os.path.join(self.cache_path, "reports.pickle"), "wb") as f:
            pickle.dump(self.reports, f)

    def set_slices(self, slices: List[Slice]):
        self.slices = dict(zip([s.slice_name for s in slices], slices))
        with open(os.path.join(self.cache_path, "slices.pickle"), "wb") as f:
            pickle.dump(self.slices, f)
